//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#ifndef __EZ_UAIL_FRAME_VECTOR_DEFINITIONS__
#define __EZ_UAIL_FRAME_VECTOR_DEFINITIONS__

//############################################################################
#include "ez_uail_frame.hpp"

//############################################################################
class ez_uail_frame_set : public vector<ez_uail_frame>
{
public:
    ez_uail_frame_set()
                     : file_name          (             ),
                       num_2D_frames      ( 0           ),
                       num_3D_frames      ( 0           ),
                       frame_index        ( 0           ),
                       frame_select_start (-1           ),
                       frame_select_end   (-1           ),
                       error              ( EZ_UAIL_OK  ),
                       null_frame         ( NULL_frame())     {}
    //------------------------------------------------------------------------
    ez_uail_frame_set(const ez_uail_frame_set& f)
                     : file_name          (f.file_name         ),
                       num_2D_frames      (f.num_2D_frames     ),
                       num_3D_frames      (f.num_3D_frames     ),
                       frame_index        (f.frame_index       ),
                       frame_select_start (f.frame_select_start),
                       frame_select_end   (f.frame_select_end  ),
                       error              (f.error             ),
                       null_frame         (NULL_frame()        )
                        {
                            reserve(f.size());
                            insert(begin(), f.begin(), f.end());
                        }
    //------------------------------------------------------------------------
    ez_uail_frame_set(string file)
                     : file_name          (file         ),
                       num_2D_frames      ( 0           ),
                       num_3D_frames      ( 0           ),
                       frame_index        ( 0           ),
                       frame_select_start (-1           ),
                       frame_select_end   (-1           ),
                       error              ( EZ_UAIL_OK  )
                        {
                            from_file(file);
                        }
    //------------------------------------------------------------------------
    ez_uail_frame_set(ez_uail_point(*F)(int, int), int points_per_frame, int num_frames);
    //------------------------------------------------------------------------
virtual
   ~ez_uail_frame_set()    {    clear();    }
    //------------------------------------------------------------------------
    void set_frame_index(int index) //sets frame_index
                    {
                        if(number_of_frames())
                        {
                            if(index >= number_of_frames())
                                index = 0;
                            if(index < 0)
                                index = number_of_frames() - 1;
                            frame_index = index;
                        }
                        else
                            frame_index = 0;
                    }
    //------------------------------------------------------------------------
    ez_uail_frame& frame(int index) //sets frame_index
                    {
                        if(number_of_frames())
                        {
                            if(index >= number_of_frames())
                                index = 0;
                            if(index < 0)
                                index = number_of_frames() - 1;
                            frame_index = index;
                            return at(frame_index);
                        }
                        return null_frame;
                    }
    //------------------------------------------------------------------------
    ez_uail_frame& next_frame() //sets frame_index
                    {
                        if(number_of_frames())
                        {
                            frame_index++;
                            if(frame_index >= number_of_frames())
                                frame_index = 0;
                            return at(frame_index);
                        }
                        return null_frame;
                    }
    //------------------------------------------------------------------------
    ez_uail_frame& previous_frame() //sets frame_index
                    {
                        if(number_of_frames())
                        {
                            frame_index--;
                            if(frame_index < 0)
                                frame_index = number_of_frames() - 1;
                            return at(frame_index);
                        }
                        return null_frame;
                    }
    //------------------------------------------------------------------------
    ez_uail_frame  frame_picker(int index) const
                    {
                        if(number_of_frames())
                        {
                            if(index < 0)
                                index = number_of_frames() - 1;
                            else if(index >= number_of_frames())
                                index = 0;
                            return at(index);
                        }
                        return null_frame;
                    }
    //------------------------------------------------------------------------
    ez_uail_frame& current_frame()
                    {
                        if(number_of_frames())
                            return at(frame_index);
                        return null_frame;
                    }
    //------------------------------------------------------------------------
    void  delete_current_frame()
            {
                if(number_of_frames() > 1)
                {
                    erase(begin() + frame_index);
                    frame_index--;
                    if(frame_index < 0)
                        frame_index = 0;
                }
                else
                {
                    clear();
                    push_back(null_frame);
                }
                return;
            }
    //------------------------------------------------------------------------
    void  delete_frame(int index)
            {
                if(number_of_frames() > 1)
                {
                    if(index < number_of_frames())
                    {
                        erase(begin() + index);
                        if(frame_index >= index)
                        {
                            frame_index--;
                            if(frame_index < 0)
                                frame_index = number_of_frames() - 1;
                        }
                    }
                }
                else
                {
                    clear();
                    push_back(null_frame);
                }
                return;
            }
    //------------------------------------------------------------------------
    void  push_back_frame(const ez_uail_frame& frame)
            {
                push_back(frame);
                return;
            }
    //------------------------------------------------------------------------
    void  count_frames()
            {
                num_2D_frames = num_3D_frames = 0;
                for(int i = 0; i < number_of_frames(); i++)
                    at(i).is_2D() ? num_2D_frames++ : num_3D_frames++;
                return;
            }
    //------------------------------------------------------------------------
    int   total_points() const
            {
                int  i,
                     points = 0;
                for(i = 0; i < number_of_frames(); i++)
                    points += (at(i).number_of_points());
                return points;
            }
    //------------------------------------------------------------------------
    ez_uail_frame& operator ++ () //pre
                    {
                        if(number_of_frames())
                        {
                            frame_index++;
                            if(frame_index >= number_of_frames())
                                frame_index = 0;
                            return at(frame_index);
                        }
                        return null_frame;
                    }
    //------------------------------------------------------------------------
    ez_uail_frame& operator -- () // pre
                    {
                        if(number_of_frames())
                        {
                            frame_index--;
                            if(frame_index < 0)
                                frame_index = number_of_frames() - 1;
                            return at(frame_index);
                        }
                        return null_frame;
                    }
    //------------------------------------------------------------------------
    ez_uail_frame& operator ++ (int = 0) //post
                    {
                        if(number_of_frames())
                        {
                            int frame_return = frame_index;
                            frame_index++;
                            if(frame_index >= number_of_frames())
                                frame_index = 0;
                            return at(frame_return);
                        }
                        return null_frame;
                    }
    //------------------------------------------------------------------------
    ez_uail_frame& operator -- (int = 0) // post
                    {
                        if(number_of_frames())
                        {
                            int frame_return = frame_index;
                            frame_index--;
                            if(frame_index < 0)
                                frame_index = number_of_frames() - 1;
                            return at(frame_return);
                        }
                        return null_frame;
                    }
    //------------------------------------------------------------------------
    ez_uail_frame_set& operator  = (const ez_uail_frame_set& ail)
                        {
                            clear();
                            reserve(ail.size());
                            insert(begin(), ail.begin(), ail.end());
                            file_name     = ail.file_name     ;
                            num_2D_frames = ail.num_2D_frames ;
                            num_3D_frames = ail.num_3D_frames ;
                            renumber();
                            return *this;
                        }
    //------------------------------------------------------------------------
    ez_uail_frame_set& operator += (const ez_uail_frame_set& ail)
                        {
                            reserve(size() + ail.size());
                            insert(end(), ail.begin(), ail.end());
                            renumber();
                            return *this;
                        }
    //------------------------------------------------------------------------
    ez_uail_frame_set& operator += (const ez_uail_frame& frame);
    //------------------------------------------------------------------------
    ez_uail_frame_set& operator += (ez_uail_segment& segment)
                        {
                            ez_uail_header header;
                            header.format   = segment.is_2D();
                            header.quantity = (u_short)segment.number_of_points();
                            push_back(ez_uail_frame(segment, header));
                            renumber();
                            return *this;
                        }
    //------------------------------------------------------------------------
    ez_uail_frame_set selected_frames()
                        {
                            int                i;
                            ez_uail_frame_set  selected;
                            for(i = 0; i < number_of_frames(); i++)
                                if(at(i).is_selected)
                                    selected += at(i);
                            return selected;
                        }
    //------------------------------------------------------------------------
    ez_uail_frame_set  operator +  (const ez_uail_frame_set& ail) const
                        {
                            ez_uail_frame_set sum(*this);
                            sum += ail;
                            return sum;
                        }
    //------------------------------------------------------------------------
    ez_uail_frame& operator [] (int index) // does not set frame_index
                    {
                        if(number_of_frames())
                        {
                            if(index >= number_of_frames())
                                index = 0;
                            if(index < 0)
                                index = number_of_frames() - 1;
                            return at(index);
                        }
                        return null_frame;
                    }
    //------------------------------------------------------------------------
    ez_uail_frame sum_of_frames()
                    {
                        bool           same_palettes = true;
                        int            i;
                        ez_uail_frame  frame(at(0).palette_index);
                        for(i = 1; i < number_of_frames(); i++)
                            if(frame.palette_index != at(i).palette_index)
                            {
                                same_palettes = false;
                                break;
                            }
                        if(!same_palettes)
                        {
                            frame.promote_to_true_color();
                            for(i = 0; i < number_of_frames(); i++)
                                at(i).promote_to_true_color();
                        }
                        for(i = 0; i < number_of_frames(); i++)
                        {
                            if(at(i).size() > 1)
                                frame += at(i);
                        }
                        return frame;
                    }
    //------------------------------------------------------------------------
    int                 number_of_frames() const {   return (int)size();   }
    //------------------------------------------------------------------------
    ez_uail_frame_set&  reverse               (           );
    ez_uail_frame_set&  reverse_selected      (           );
    ez_uail_frame_set&  random_order          (           );
    ez_uail_frame_set&  frame_reverse         (           );
    ez_uail_frame_set&  superimpose_from_file (string file);
    //------------------------------------------------------------------------
    int     from_file                 (string    file,
                                       bool      append = false)       ;
    //------------------------------------------------------------------------
    int     from_ifstream_uail        (ifstream& in,
                                       bool      append = false)       ;
    //------------------------------------------------------------------------
    int     from_wave_file            (string    file,
                                       bool      append = false)       ;
    //------------------------------------------------------------------------
    int     from_ADAT_wave_files      (string    file_1,
                                       string    file_2,
                                       string    file_3,
                                       string    file_4        )       ;
    //------------------------------------------------------------------------
    int     from_ifstream_wave        (ifstream& in,
                                       bool      append = false)       ;
    //------------------------------------------------------------------------
    int     measure_ADAT_waves        (//ifstream& in_1,
                                       ifstream& in_2,
                                       ifstream& in_3,
                                       ifstream& in_4          )       ;
    //------------------------------------------------------------------------
    int     from_ifstreams_ADAT_waves (ifstream& in_1,
                                       ifstream& in_2,
                                       ifstream& in_3,
                                       ifstream& in_4          )       ;
    //------------------------------------------------------------------------
    int     from_dxf_file             (string    file,
                                       bool      append = false)       ;
    //------------------------------------------------------------------------
    int     from_dxf_directory        (string    dir,
                                       bool      append = false)       ;
    //------------------------------------------------------------------------
    int     from_ifstream_dxf         (ifstream& in,
                                       bool      append = false)       ;
    //------------------------------------------------------------------------
    int     from_shp_file             (string    file          )       ;
//    int     from_ifstream_shp         (ifstream& in            )       ;
    void    flatten_z                 (                        )       ;
    void    z_order_points            (unsigned short span     )       ;
    void    unblank_all_points        (                        )       ;
    void    convert_black_to_blank    (                        )       ;
    void    minimize_blank_points     (                        )       ;
    void    remove_dwell_points       (                        )       ;
    void    minimize_vectors          (                        )       ;
    void    minimize_points           (                        )       ;
    void    accentuate_the_positive   (                        )       ;
    void    eliminate_the_negative    (                        )       ;
    void    explode_current_frame     (                        )       ;
    void    explode_all_frames        (                        )       ;
    void    collapse_selected_frames  (                        )       ;
    void    delete_every_nth_frame    (unsigned short n        )       ;
    void    fracture_segments         (                        )       ;
    void    blank_equivelant_segments (                        )       ;
    void    bond_segments             (                        )       ;
    void    reorder_segments          (                        )       ;
    void    add_dwell_points          (                        )       ;
    void    add_lit_span_points       (                        )       ;
    void    add_blank_span_points     (                        )       ;
    void    delete_redundant_frames   (                        )       ;
    void    optimize_points           (                        )       ;
    void    rainbow_recolor           (int effect, float factor)       ;
    void    quarter_turn              (int plane, int turns    )       ;
    int     scale_around_origin       (ez_3D_float f           )       ;
    int     scale_around_origin       (                        )       ;
    int     rotate_around_origin      (ez_3D_float f           )       ;
    int     rotate_around_origin      (                        )       ;
    int     move                      (ez_3D_float f           )       ;
    int     move                      (                        )       ;
    void    toggle_frame_select_start (                        )       ;
    void    toggle_frame_select_end   (                        )       ;
    void    delete_selected_frames    (                        )       ;
    void    trim_to_selected_frames   (                        )       ;
    void    bit_reduce_to_palette     (                        )       ;
    void    best_reduce_to_palette    (                        )       ;
    void    promote_to_true_color     (                        )       ;
    void    best_match_palette        (int index               )       ;
    void    best_match_target_palette (                        )       ;
    void    strip_color               (                        )       ;
    void    accept_palette            (int index               )       ;
    void    accept_target_palette     (                        )       ;
    void    color_over_ezfb_bitmap    (string file             )       ;
    void    color_over_ezfb_bitmap    (struct ezfb_bitmap* bmp )       ;
    void    flip                      (int plane               )       ;
    void    renumber                  (                        )       ;
    void    save_as_uail              (string file             )       ;
    void    to_ofstream_uail          (ofstream& out           )       ;
    void    display_stats             (                        )       ;
    int     total_wave_samples        (                        ) const ;
    void    save_as_wave              (string file             ) const ;
    void    to_ofstream_wave          (ofstream& out           ) const ;
    void    save_as_bmp_directory     (string dir              ) const ;
    void    save_as_dxf_directory     (string dir              ) const ;
    void    save_as_code              (string object_name      )       ;
    //------------------------------------------------------------------------
    string              file_name         ;
    
    int                 num_2D_frames     ,
                        num_3D_frames     ,
                        frame_index       ,
                        frame_select_start,
                        frame_select_end  ,
                        error             ;
                        
    float               max_r             , // for ADAT wave import
                        max_g             ,
                        max_b             ;
                        
    ez_uail_frame       null_frame        ;
    
    ez_uail_segment     intro             ;
};

//############################################################################
typedef void(*ez_uail_frame_effect    )(ez_uail_frame*    );
typedef void(*ez_uail_frame_set_effect)(ez_uail_frame_set*);

//############################################################################
#endif

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
