//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#ifndef __EZ_UAIL_SEGMENT_DEFINITIONS__
#define __EZ_UAIL_SEGMENT_DEFINITIONS__

//############################################################################
#include "ez_bmp.hpp"
#include "ez_real_segment.hpp"

//############################################################################
class ez_uail_frame_set; //predeclaration

//############################################################################
class ez_uail_segment : public vector<ez_uail_point>
{
public:
    ez_uail_segment()
                     : palette_index (EZ_UAIL_DEFAULT),
                       error         (EZ_UAIL_OK     )      {}
    //------------------------------------------------------------------------
    ez_uail_segment(const int& palette_index) // empty segment
                     : palette_index (palette_index),
                       error         (EZ_UAIL_OK   )        {}
    //------------------------------------------------------------------------
    ez_uail_segment(const int& palette_index, int not_used) // not empty segment
                     : palette_index (palette_index),
                       error         (EZ_UAIL_OK   )
                    {
                        reserve(2);
                        push_back(ez_uail_point());
                        push_back(ez_uail_point());
                    }
    //------------------------------------------------------------------------
    ez_uail_segment(const ez_uail_segment& segment)
                     : palette_index (segment.palette_index),
                       error         (EZ_UAIL_OK           )
                    {
                        clear();
                        reserve(segment.size());
                        insert(end(), segment.begin(), segment.end());
                    }
    //------------------------------------------------------------------------
    ez_uail_segment(ez_uail_point from, ez_uail_point to);
    //------------------------------------------------------------------------
    ez_uail_segment(ez_uail_point p1,
                    ez_uail_point p2,
                    float         rotation,
                    string        font,
                    string        text
                   );   // 3D text constructor
    //------------------------------------------------------------------------
    ez_uail_segment(ez_uail_point center,
                    ez_uail_point arc_start,
                    float         arc_angle // OK degrees, whatever!
                   );   // 3D circular arc constructor
    //------------------------------------------------------------------------
    ez_uail_segment(ez_uail_point center,
                    ez_uail_point arc_start,
                    float         arc_angle,
                    float         radii_ratio
                   );   // 3D elliptical arc constructor
    //------------------------------------------------------------------------
    ez_uail_segment(ez_uail_point center,
                    ez_uail_point first_vertex,
                    int           number_of_sides
                   );   // 3D polygon constructor
    //------------------------------------------------------------------------
   ~ez_uail_segment()    {    clear();    };
    //------------------------------------------------------------------------
    bool operator == (const ez_uail_segment& segment)
            {
                if(palette_index != segment.palette_index)
                    return false;
                if(size() != segment.size())
                    return false;
                for(int i = 0; i < number_of_points(); i++)
                    if(at(i) != segment.at(i))
                        return false;
                return true;
            }
    //------------------------------------------------------------------------
    ez_uail_segment&  operator += (const ez_uail_segment& segment)
                        {
                            if(palette_index != segment.palette_index)
                                promote_to_true_color();
                            reserve(size() + segment.size());
                            insert(end(), segment.begin(), segment.end());
                            return *this;
                        }
    //------------------------------------------------------------------------
    ez_uail_segment&  operator += (const ez_uail_point& point)
                        {
                            push_back(point);
                            return *this;
                        }
    //------------------------------------------------------------------------
    ez_uail_segment&  operator =  (const ez_uail_segment& segment)
                        {
                            clear();
                            //palette_index = segment.palette_index;
                            reserve(segment.size());                            
                            insert(begin(), segment.begin(), segment.end());
                            return *this;
                        }
    //------------------------------------------------------------------------
    bool is_2D() const
            {
                if(number_of_points() > 1)
                {
                    for(int i = 0; i < number_of_points(); i++)
                        if(at(i).z != 0)
                            return false;
                }
                return true;
            }
    //------------------------------------------------------------------------
    ez_uail_point  first_lit_point() const
                    {
                        for(int i = 1; i < number_of_points(); i++)
                            if(at(i).is_lit())
                                return at(i);
                        return ez_uail_point(0);
                    }
    //------------------------------------------------------------------------
    int number_of_points() const 
            {
                return (int)size();
            }
    //------------------------------------------------------------------------
    int first_lit_point_index() const
            {
                for(int i = 1; i < number_of_points(); i++)
                    if(at(i).is_lit())
                        return i;
                return -1;
            }
    //------------------------------------------------------------------------
    ez_uail_point first_lit_anchor() const
                    {
                        for(int i = 1; i < number_of_points(); i++)
                            if(at(i).is_lit())
                                return at(i - 1);
                        return ez_uail_point(0);
                    }
    //------------------------------------------------------------------------
    int first_lit_anchor_index() const
            {
                for(int i = 1; i < number_of_points(); i++)
                    if(at(i).is_lit())
                        return (i - 1);
                return -1;
            }
    //------------------------------------------------------------------------
    ez_uail_point last_lit_anchor() const
                    {
                        for(int i = number_of_points() - 1; i > 0; i--)
                            if(at(i).is_lit())
                                return at(i - 1);
                        return ez_uail_point(0);
                    }
    //------------------------------------------------------------------------
    ez_uail_point last_lit_point() const
                    {
                        for(int i = number_of_points() - 1; i > 0; i--)
                            if(at(i).is_lit())
                                return at(i);
                        return ez_uail_point(0);
                    }
    //------------------------------------------------------------------------
    int last_lit_point_index() const
            {
                for(int i = number_of_points() - 1; i > 0; i--)
                    if(at(i).is_lit())
                        return i;
                return -1;
            }
    //------------------------------------------------------------------------
    int number_of_lit_points() const
            {
                int count = 0;
                for(int i = 1; i < number_of_points(); i++)
                    if(at(i).is_lit())
                        count++;
                return count;
            }
    //------------------------------------------------------------------------
    int number_of_blank_points() const
            {
                int count = 1; //first point is always an anchor
                for(int i = 1; i < number_of_points(); i++)
                    if(at(i).is_blank())
                        count++;
                return count;
            }
    //------------------------------------------------------------------------
    int number_of_unique_colors() const
            {
                if(number_of_points() > 1)
                {
                    int         i, j;
                    ez_palette  palette;
                    palette.push_back((ez_color)first_lit_point());                    
                    for(i = first_lit_point_index() + 1; i < number_of_points(); i++)
                    {
                        for(j = 0; j < palette.number_of_colors(); j++)
                        {
                            if(palette[j] == (ez_color)at(i))
                                break;
                        }
                        if(j == palette.number_of_colors())
                            palette.push_back(at(i));
                    }
                    return palette.number_of_colors();
                }
            }
    //------------------------------------------------------------------------
    ez_palette as_ez_palette() const
            {
                ez_palette palette(number_of_points());
                for(int i = 0; i < number_of_points(); i++)
                    palette[i] = (ez_color)(at(i));
                return palette;
            }
    //------------------------------------------------------------------------
    ez_uail_segment& remove_point(int point_index)
            {
                if(    point_index >= 0
                    && point_index <  (int)size()
                    && size()
                  )
                {
                    if(size() == 1)
                        clear();
                    int              i;
                    ez_uail_segment  segment;
                    segment.reserve(size());
                    for(i = 0; i < point_index; i++)
                        segment += at(i);
                    for(i = point_index + 1; i < (int)size(); i++)
                        segment += at(i);
                    *this = segment;
                }
                return *this;
            }
    //------------------------------------------------------------------------
    ez_3D_short       segment_front               (                               ) const ;
    ez_3D_short       segment_back                (                               ) const ;
    ez_3D_short       segment_top                 (                               ) const ;
    ez_3D_short       segment_bottom              (                               ) const ;
    ez_3D_short       segment_right               (                               ) const ;
    ez_3D_short       segment_left                (                               ) const ;
    short             segment_height              (                               ) const ;
    short             segment_width               (                               ) const ;
    short             segment_depth               (                               ) const ;
    short             segment_size                (                               ) const ;
    bool              is_closed_polygon           (                               ) const ;
    ez_uail_segment&  reverse                     (                               )       ;
    ez_uail_segment   shortest_path_of_segments   (                               )       ;
    void              blank_all_points            (                               )       ;
    void              unblank_all_points          (                               )       ;
    //void              remove_added_points         (                               )       ;
    void              minimize_blank_points       (                               )       ;
    void              remove_dwell_points         (                               )       ;
    void              blank_equivelant_segments   (                               )       ;
    void              minimize_vectors            (                               )       ;
    void              minimize_points             (                               )       ;    
    void              accentuate_the_positive     (                               )       ;
    void              eliminate_the_negative      (                               )       ;
    ez_uail_segment&  reorder_segments            (                               )       ;
    ez_uail_segment&  fracture_segments           (                               )       ;
    ez_uail_segment&  bond_segments               (                               )       ;
    ez_uail_segment&  add_new_point               (int  point_index               )       ;
    ez_uail_segment&  break_segment               (int& point_index               )       ;
    ez_uail_segment&  connect_the_dots            (int p1, int p2                 )       ;
    double            next_point_angle            (int point                      ) const ;
    double            total_angle                 (                               ) const ;
    double            max_angle                   (                               ) const ;
    double            next_point_distance         (int point                      ) const ;
    double            total_distance              (                               ) const ;
    double            max_distance                (                               ) const ;
    void              add_dwell_points            (                               )       ;
    void              add_lit_span_points         (                               )       ;
    void              add_blank_span_points       (                               )       ;
    void              flip                        (int plane                      )       ;
    void              quarter_turn                (int plane, int turns           )       ;
    void              z_order_points              (unsigned short span            )       ;
    void              flatten_z                   (                               )       ;
    int               rotate                      (ez_3D_float a                  )       ;
    int               rotate_around_origin        (ez_3D_float a                  )       ;
    int               rotate_around_point         (ez_3D_float p, ez_3D_float a   )       ;
    int               rotate_around_point_x       (ez_3D_float p, float a         )       ;
    int               rotate_around_point_y       (ez_3D_float p, float a         )       ;
    int               rotate_around_point_z       (ez_3D_float p, float a         )       ;
    int               rotate_around_origin        (                               )       ;
    void              ripple                      (int   direction,
                                                   float amplitude,
                                                   float freq     ,
                                                   float phase                    )       ;
    //------------------------------------------------------------------------                                                   
    ez_palette        to_color_table              (                               ) const ;
    void              strip_color                 (                               )       ;
    void              accept_palette              (int index                      )       ;
    void              accept_target_palette       (                               )       ;
    void              best_match_palette          (int index                      )       ;
    void              best_match_target_palette   (                               )       ;
    void              convert_black_to_blank      (                               )       ;
    void              rainbow_recolor             (int effect, float factor       )       ;
    void              rotate_colors               (                               )       ;
    bool              find_rgb_in_palette         (const ez_palette& palette      )       ;
    void              set_rgb_from_palette        (                               )       ;
    void              set_palette_to_332          (                               )       ;
    void              sync_rgb_and_palette        (                               )       ;
    void              bit_reduce_to_palette       (                               )       ;
    void              best_reduce_to_palette      (                               )       ;
    void              promote_to_true_color       (                               )       ;
    void              color_over_ezfb_bitmap      (string file                    )       ;
    void              color_over_ezfb_bitmap      (struct ezfb_bitmap* bmp        )       ;
    //------------------------------------------------------------------------
    int               move                        (ez_3D_float d                  )       ;
    int               move                        (                               )       ;
    int               scale                       (ez_3D_float s                  )       ;
    int               scale_around_point          (ez_3D_float p, ez_3D_float s   )       ;
    int               scale_around_origin         (ez_3D_float s                  )       ;
    int               scale_around_origin         (                               )       ;
    ez_3D_float       find_center                 (                               ) const ;
    int               number_of_segments          (                               ) const ;
    bool              find_segment                (int index, int& start, int& end) const ;
    bool              find_segment                ( int point_index,
                                                    int& start,
                                                    int& end,
                                                    int& segment_index
                                                  )                                 const ;
    ez_uail_segment   find_segment                (int index                      ) const ;
    ez_uail_frame_set find_segments               (                               ) const ;
    ez_3D_float       find_segment_center         (int index                      ) const ;
    int               segment_index_of_point      (int index                      ) const ;
    int               move_segment                (int index, ez_3D_float d       )       ;
    int               rotate_segment              (int index, ez_3D_float a       )       ;
    int               rotate_segment_around_origin(int index, ez_3D_float a       )       ;
    int               scale_segment               (int index, ez_3D_float m       )       ;
    int               scale_segment_around_origin (int index, ez_3D_float m       )       ;
    //------------------------------------------------------------------------
    void              to_ofstream_wave            (ofstream&  out,
                                                    const bool& end_of_frame = false,
                                                    const bool& unique_frame = false
                                                  )                                 const ;
    //------------------------------------------------------------------------
    int               palette_index,
                      error        ;
    //------------------------------------------------------------------------
};

//############################################################################
ez_uail_segment blank_segment();
ez_uail_segment NULL_segment ();

//############################################################################
#endif

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
